/* 
 *  $Id: DLClassLoader.java,v 1.3 1997/08/28 18:34:41 djw Exp $
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */


package ants;

import java.util.Hashtable;


/**
 * Demand Loading ClassLoader.  Presents an interface to load a
 * code group from byte arrays received by the DL protocol or 
 * registered at the Node.
 *
 * @author David Wetherall
 */


// warning -- there be wicked interactions here.  this code is 
// heavily commented and asserted as a result. see debug also.

final class DLClassLoader extends ClassLoader {

  /**
   * Cache of classes loaded by all local DLClassLoaders
   * into this java vm.  Used to avoid multiple definitions.
   */

  private static Hashtable sharedRegistry = new Hashtable();

  /**
   * Cache of classes loaded by this DLClassLoader. Used to
   * determine what is accessibile in the capsule loading space.
   */

  private Hashtable registry;

  /** Side information for loadClass -- the code group being loaded */
  private CodeGroup inProgress;

  /** Side information for loadClass -- local registration or DL protocol */
  private boolean isDL;

  
  /** Compile time flag for debugging statements */
  private static boolean DEBUG = false;

  /** Debug print statement, redirects to logger */
  private static void debug(String msg) {
    if (DEBUG) System.err.println("DLClassLoader: "+ msg);
  }


  /**
   * This method is the catch-all to intercept any unknown class 
   * references that are needed by capsules.  Eventually all classes
   * should be found already in the registry, or an error occurs.
   */

  protected synchronized Class loadClass(String name, boolean resolve) 
    throws ClassNotFoundException
  {
    Entity.assert(sharedRegistry.get(name) != this,
		  "DLClassLoader: recursive load of "+ name);
    
    if (inProgress != null) {
      debug("loadClass: called by codegroup for "+ name);
    } else {
      debug("loadClass: called by system for "+ name);
    }
    
    Class cl = (Class)registry.get(name);
    if (cl != null) {
      debug("loadClass: result found in registry for "+ name);
    }

    if ((inProgress != null) && (cl == null)) {
      CodeGroup g = inProgress;

      // this is a group load, not already in the registry,
      // so it must be another class in the code group ...

      for (int i = 0; i < g.meths.size(); i++) {
	Method m = (Method)g.meths.elementAt(i);

	// first, we find the right method
	if (name.equals(m._cn)) {

	  // got it, see if the result is known
	  cl = (Class)sharedRegistry.get(name);
	  if (cl != null) {
	    debug("loadClass: result found in shared registry for "+ name);
	    break;
	  }

	  // if not, use a placeholder to bootstrap the origin check
	  sharedRegistry.put(name, this);	  
	  if (isDL) {
	    debug("loadClass: calling defineClass for "+ name);
	    cl = defineClass(m.code, 0, m.code.length);

	    // about the only thing we can do is check afterwards
	    // in the hope of catching prior definitions

	    Entity.assert(cl.getClassLoader() == this,
			  "DLClassLoader: multiple definitions for "+ name);
	  } else {
	    debug("loadClass: finding via class.forname for "+ name);
	    cl = Class.forName(name);
	  }

	  // we've tried all cases, done
	  break;
	}
      }
    }

    // if we still don't have it, then breach the load boundary
    if (cl == null) {
      Entity.warn("DLClassLoader: breach load for "+ name);

      cl = (Class)sharedRegistry.get(name);
      if (cl != null) {
	debug("loadClass: result found in shared registry for "+ name);
      } else {
	debug("loadClass: finding via class.forname for "+ name);
	cl = findSystemClass(name);
      }
    }
    
    Entity.assert(cl != null, "DLClassLoader: empty load for "+ name);
    Entity.assert(cl.getName().equals(name),
		  "DLClassLoader: unexpected name ("+ 
		  cl.getName() +") for "+ name);
    
    // update both registries with the right info
    registry.put(name, cl);
    sharedRegistry.put(name, cl);
    
    if (resolve) {
      debug("loadClass: calling resolveClass for"+ name);
      resolveClass(cl);
    }
    
    debug("loadClass: returning for "+ name);
    return cl;
  }


  /**
   * Entry point used by a node to register code groups.
   * Loads ultimately come from the local filesystem
   */

  synchronized boolean loadForRegister(CodeGroup g) 
    throws Exception
  {
    Entity.assert(inProgress == null, "DLClassLoader: Interleaved load");
    
    debug("loadForRegister: begin loading "+ g);

    try {
      inProgress = g;
      isDL = false;
      
      for (int i = 0; i < g.meths.size(); i++) {
	Method m = (Method)g.meths.elementAt(i);
	
	// load failures will throw an exception here, which
	// should percolate up to the application registering
	m.cl = loadClass(m._cn, true);
      }
    } catch (Exception e) {
      throw e;
    } finally {

      // must release
      inProgress = null;
    }

    debug("loadForRegister: end loading "+ g);
  
    return true;
  }


  /**
   * Entry point used by a the DL protocol load code groups.
   * Loads ultimately come from byte arrays.
   */

  synchronized boolean loadOnDemand(CodeGroup g) {
    Entity.assert(inProgress == null, "DLClassLoader: Interleaved load");
    
    debug("loadOnDemand: begin loading "+ g);

    boolean loaded = true;
    try {
      inProgress = g;
      isDL = true;
      
      for (int i = 0; i < g.meths.size(); i++) {
	Method m = (Method)g.meths.elementAt(i);
	
	debug("loadOnDemand: trial loading "+ m._cn);
	
	try {
	  loadClass(m._cn, true);
	} catch (ClassNotFoundException e) {
	  loaded = false;
	  break;
	}
      }
      
      if (loaded) {
	for (int i = 0; i < g.meths.size(); i++) {
	  Method m = (Method)g.meths.elementAt(i);
	  
	  debug("loadOnDemand: final loading "+ m._cn);

	  try {
	    m.cl = loadClass(m._cn, true);
	  } catch (ClassNotFoundException e) {
	    Entity.error("DLClassLoader: unexpected failure to load "+ m._cn);
	  }
	}
      }
    } finally {
      
      // must release
      inProgress = null;
    }
    
    debug("loadOnDemand: end loading "+ g);
    return loaded;
  }
  

  /**
   * Export node and extension classes that capsules may access
   * by adding them to the registry.
   */

  public synchronized void exportClass(String name) 
    throws ClassNotFoundException
  {

    debug("exportClass: adding to registry for "+ name);
 
    // all exported classes must be locally loadable
    Class cl = Class.forName(name);
    registry.put(name, cl);
    sharedRegistry.put(name, cl);
  }
  
  /** Convenience callback for capsules to check their load path */
  public static void verifyOrigin(String name) {
    if (sharedRegistry.get(name) == null)
      Entity.error("DLClassLoader: "+ name +" fails load origin check");
  }
  
  /** One DLClassLoader is constructed by each Node */
  DLClassLoader() {
    registry = new Hashtable();
    inProgress = null;
    isDL = false;
  }
}



